import logging
from logging import Handler, LogRecord
from typing import Any, List, Optional

from blessings import Terminal


class Simple(logging.Formatter):
    """
    Fancy, colorised log output adding ANSI escape codes to the log output.

    :params show_threads: Whether to display thread names or not.
    :params show_exc: Whether to display tracebacks or not.

    .. note:: This formatter *suppresses* tracebacks by default! Remember that
        is is meant to give a concise, readable output. If you need to see
        tracebacks on the console, you can override this setting useing
        \*show_exc\*.
    """

    @staticmethod
    def basicConfig(show_exc=True, show_threads=False, **kwargs):
        # type: (bool, bool, Any) -> List[Handler]
        '''
        Convenience method to have a one-liner set-up.

        Both *show_exc* and *show_threads* are directly passed to
        :py:class:`~.Simple`. The remaining *args* and *kwargs* are passed on to
        :py:func:`logging.basicConfig`.

        After returning from :py:func:`logging.basicConfig`, it will fetch the
        *stderr* and *stdout* handlers and replace the formatter.

        The function also returns a list of all handlers which have been
        modified. This is useful if you want to modify the handlers any further
        (for example using :py:class:`~gouge.filters.ShiftingFilter`).
        '''
        logging.basicConfig(**kwargs)
        root = logging.getLogger()
        output = []
        for handler in root.handlers:
            stream = getattr(handler, 'stream', None)
            if not stream:
                continue

            stream_name = getattr(stream, 'name', None)
            if stream_name not in ('<stderr>', '<stdout>'):
                continue

            handler.setFormatter(Simple(show_exc, show_threads))
            output.append(handler)
        return output

    def __init__(self, show_exc=False, show_threads=False, fmt=None,
                 datefmt=None):
        # type: (bool, bool, Optional[str], Optional[str]) -> None
        logging.Formatter.__init__(self, fmt, datefmt)
        self.show_threads = show_threads
        self.show_exc = show_exc
        self.term = Terminal()

    def format(self, record):
        # type: (LogRecord) -> str
        if record.levelno <= logging.DEBUG:
            colorize = self.term.bold_black
        elif record.levelno <= logging.INFO:
            colorize = self.term.cyan
        elif record.levelno <= logging.WARNING:
            colorize = self.term.yellow
        elif record.levelno <= logging.ERROR:
            colorize = self.term.bold_red
        else:
            colorize = self.term.bold_yellow_on_red

        record.message = record.getMessage()
        record.asctime = self.formatTime(record, self.datefmt or '')

        message_items = [
            '{t.green}{asctime}{t.normal}',
            '{levelcolor}{levelname:<10}{t.normal}',
            '{t.bold}[{name}]{t.normal}',
            '{message}',
        ]
        if self.show_threads:
            message_items.insert(0, '{threadName:<10}')

        message_template = ' '.join(message_items)

        if self.show_exc:
            if record.exc_info:
                # Cache the traceback text to avoid converting it multiple times
                # (it's constant anyway)
                exc_text = getattr(record, 'exc_text', '')
                if not exc_text:
                    record.exc_text = self.formatException(  # type: ignore
                        record.exc_info)

            exc_text = getattr(record, 'exc_text', '')
            if exc_text:
                message_template += '\n{t.red}{exc_text}{t.normal}'

        return message_template.format(
            t=self.term,
            levelcolor=colorize,
            **vars(record))



import logging

import pytest
from gouge.colourcli import Simple


def test_caplog(caplog):
    '''
    Using the pytest "caplog" fixture can cause trouble with the "Simple"
    formatter. This test uses the fixture, and should not cause any errors.
    '''
    Simple.basicConfig(level=logging.DEBUG)
    logging.warning('Hello World')
    assert len(caplog.records) == 1



"""
This module contains generally useful filters.
"""
import logging
from typing import Set
from warnings import warn


class ShiftingFilter:
    """
    This filter will shift the logging level of log records a certain number of
    log levels.

    For example:

    ``ShiftingFilter(1)`` will convert a message with level ``INFO`` to
    ``WARNING``, and ``ShiftingFilter(-1)`` will do the reverse.

    An example use-case is controlling log levels of libraries: It is possible
    that a library emits error-messages, but in the context of your application
    you would like to see those as *warning* messages only. If you have no
    control over that library your hands are tied and the only thing you can do
    is suppress these messages by changing the level of the associated logger.
    By modifying the log level using this filter still allows you to see the
    messages, albeit with another level which is more appropriate to your
    application.

    .. note::
        Filters can either be attached to handlers or loggers. But be aware that
        when attaching to loggers, they will only trigger on the exact logger
        they were attached to. Not in the parent hierarchy! See
        :py:meth:`logging.Filter.filter` for details. To simplify attaching to
        loggers, this class offers the method
        :py:meth:`~.ShiftingFilter.inject`.

    :param shift_by: The number of levels to shift. Positive integers will
        increase the level, negative integers will decrease it.
    :param logger: This and all child loggers will be shifted.
    :param min: Don't shift below this level.
    :param max: Don't shift above this level.
    :param offset: An explicit, fine-grained offset value. This overrides
        *shift_by*!
    """
    def __init__(self, shift_by=0, logger='',
                 min=logging.NOTSET, max=logging.CRITICAL, offset=0):
        # type: (int, str, int, int, int) -> None
        if offset and shift_by:
            warn('You specified both "offset" and "shift_by"! "shift_by" '
                 'will be ignored!', SyntaxWarning)

        if offset:
            self.offset = offset
        else:
            self.offset = 10 * shift_by
        self.logger = logger
        self.max = max
        self.min = min
        self.injected_loggers = set()  # type: Set[logging.Logger]

    def inject(self, parent):
        # type: (str) -> None
        """
        Loop over each known logger and attach this filter.

        You can remove the attached filters again using
        :py:meth:`~.ShiftingFilter.cleanup`.

        .. note::
            This will only attach the filter to loggers which already exist! If
            you see that something is not working as expected, make sure the
            logger exists before calling this! You can look at all the existing
            loggers using ``logging.Logger.manager.loggerDict``.

        :param parent: Attach the filter to this and all descendant loggers.
        """
        if not isinstance(parent, str):
            parent = parent.name
        items = logging.Logger.manager.loggerDict.items()  # type: ignore
        for name, logger in items:
            if name.startswith(parent) and not isinstance(logger, logging.PlaceHolder):
                logger.addFilter(self)
                self.injected_loggers.add(logger)

    def cleanup(self):
        # type: () -> None
        """
        Remove all filters applied via :py:meth:`~.ShiftingFilter.inject`.
        """
        for logger in self.injected_loggers:
            logger.removeFilter(self)  # type: ignore

    def filter(self, record):
        # type: (logging.LogRecord) -> bool
        """
        Always returns *True* but will modify the logging level of *record* by
        the rules defined in this filter.

        See :py:meth:`logging.Filter.filter`
        """
        if record.name.startswith(self.logger):
            new_value = record.levelno + self.offset
            new_levelno = min(self.max, max(self.min, new_value))
            record.levelname = logging.getLevelName(new_levelno)
            record.levelno = new_levelno
        return True



import io
import logging
import unittest
from functools import partial

from gouge.filters import ShiftingFilter


SimpleRecord = partial(
    logging.LogRecord,
    pathname='',
    lineno=0,
    msg='',
    args={},
    exc_info=None)


class TestShiftingFilter(unittest.TestCase):

    def test_noshift(self):
        '''
        A default instance should not shift anything
        '''
        record = SimpleRecord(name='a.b.c', level=logging.DEBUG)
        filter_ = ShiftingFilter()
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.DEBUG)

    def test_noshift_unmatching(self):
        '''
        If the name does not match, no shifting should occur.
        '''
        filter_ = ShiftingFilter(1, 'foo.bar')
        record = SimpleRecord(name='a.b.c', level=logging.DEBUG)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.DEBUG)

    def test_noshift_parent(self):
        '''
        Parent nodes should not be shifted.
        '''
        filter_ = ShiftingFilter(1, 'a.b.c')
        record = SimpleRecord(name='a.b', level=logging.DEBUG)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.DEBUG)

    def test_shift_fullmatch(self):
        '''
        If the name matches exactly, we should shift
        '''
        filter_ = ShiftingFilter(1, 'a.b.c')
        record = SimpleRecord(name='a.b.c', level=logging.DEBUG)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.INFO)

    def test_shift_child(self):
        '''
        Child nodes should be shifted
        '''
        filter_ = ShiftingFilter(1, 'a.b.c')
        record = SimpleRecord(name='a.b.c.d.e', level=logging.DEBUG)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.INFO)

    def test_shift_reversed(self):
        '''
        We should also allow shifting in reverse.
        '''
        filter_ = ShiftingFilter(-1)
        record = SimpleRecord(name='a.b.c', level=logging.INFO)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.DEBUG)

    def test_shift_fine_grained(self):
        '''
        Instead of shifting by levels, the user should be allowed to shift by an
        exact offset.
        '''
        filter_ = ShiftingFilter(offset=2)
        record = SimpleRecord(name='a.b.c', level=logging.NOTSET)
        filter_.filter(record)
        self.assertEqual(record.levelno, 2)

    def test_shift_min(self):
        '''
        We should never go below level.NOTSET
        '''
        filter_ = ShiftingFilter(-1)
        record = SimpleRecord(name='a.b.c', level=logging.NOTSET)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.NOTSET)

    def test_shift_max(self):
        '''
        We also stop above logging.CRITICAL
        '''
        filter_ = ShiftingFilter(1)
        record = SimpleRecord(name='a.b.c', level=logging.CRITICAL)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.CRITICAL)

    def test_shift_custom_min(self):
        '''
        The user should have the option to refuse shifting below a certain
        level.
        '''
        filter_ = ShiftingFilter(-1, min=logging.INFO)
        record = SimpleRecord(name='a.b.c', level=logging.INFO)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.INFO)

    def test_shift_custom_max(self):
        '''
        The user should have the option to refuse shifting above a certain
        level.
        '''
        filter_ = ShiftingFilter(1, max=logging.ERROR)
        record = SimpleRecord(name='a.b.c', level=logging.ERROR)
        filter_.filter(record)
        self.assertEqual(record.levelno, logging.ERROR)

    def test_shift_custom_min_outside_bounds(self):
        '''
        We allow specifying a custom min value outside of the "standard" bounds.
        '''
        filter_ = ShiftingFilter(-1, min=-10)
        record = SimpleRecord(name='a.b.c', level=logging.NOTSET)
        filter_.filter(record)
        self.assertEqual(record.levelno, -10)

    def test_shift_custom_max_outside_bounds(self):
        '''
        We allow specifying a custom max value outside of the "standard" bounds.
        '''
        filter_ = ShiftingFilter(1, max=60)
        record = SimpleRecord(name='a.b.c', level=logging.CRITICAL)
        filter_.filter(record)
        self.assertEqual(record.levelno, 60)


class TestLoggingWithFilter(unittest.TestCase):

    def tearDown(self):
        logging.shutdown()
        for logger in logging.Logger.manager.loggerDict.values():
            if isinstance(logger, logging.PlaceHolder):
                continue
            del logger.handlers[:]
            del logger.filters[:]

    def test_attached_to_handler(self):
        blob = io.StringIO()
        handler = logging.StreamHandler(blob)
        handler.setFormatter(logging.Formatter(
            u'%(levelno)s %(levelname)s %(msg)s'))
        handler.addFilter(ShiftingFilter(-1))
        logger_c = logging.getLogger('a.b.c')
        logger_d = logging.getLogger('a.b.d')
        logger_parent = logging.getLogger('a')
        logger_parent.setLevel(logging.DEBUG)
        logger_parent.addHandler(handler)

        logger_c.error(u'error - c')
        logger_d.info(u'info - d')
        logger_d.debug(u'debug - d')

        lines = blob.getvalue().splitlines()

        expected = [
            u'30 WARNING error - c',
            u'10 DEBUG info - d',
            u'0 NOTSET debug - d',
        ]
        self.assertEqual(lines, expected)

    def test_attached_to_logger(self):
        blob = io.StringIO()
        handler = logging.StreamHandler(blob)
        handler.setFormatter(logging.Formatter(
            u'%(levelno)s %(levelname)s %(msg)s'))

        logger_c = logging.getLogger('a.b.c')
        logger_d = logging.getLogger('a.b.d')
        logger_parent = logging.getLogger('a')
        logger_parent.setLevel(logging.DEBUG)
        logger_parent.addHandler(handler)

        filter = ShiftingFilter(-1)
        filter.inject('a')

        logger_c.error(u'error - c')
        logger_d.info(u'info - d')
        logger_d.debug(u'debug - d')

        filter.cleanup()

        lines = blob.getvalue().splitlines()

        expected = [
            u'30 WARNING error - c',
            u'10 DEBUG info - d',
            u'0 NOTSET debug - d',
        ]
        self.assertEqual(lines, expected)
